#ifndef CAFFE_HASH_DATA_LAYER_HPP_
#define CAFFE_HASH_DATA_LAYER_HPP_

#include <string>
#include <vector>

#include "caffe/blob.hpp"
#include "caffe/layer.hpp"
#include "caffe/proto/caffe.pb.h"

#include "caffe/layers/base_data_layer.hpp"
#include "caffe/util/HashData.h"

#include "caffe/internal_thread.hpp"
#include "caffe/util/blocking_queue.hpp"
namespace caffe {


template <typename Dtype>
class HierHash_Label_HashLabel_Batch 
{
public:
	~HierHash_Label_HashLabel_Batch()
	{
		for (int i = 0; i<(int)m_vpHierHashes.size(); i++)
		{
			delete m_vpHierHashes[i];
		}
		m_vpHierHashes.resize(0);

		for (int i = 0; i<(int)m_vpHashLabels.size(); i++)
		{
			delete m_vpHashLabels[i];
		}
		m_vpHashLabels.resize(0);
	}
	std::vector<CHierarchyHash *> m_vpHierHashes;
	Blob<Dtype> m_label_blob;
	std::vector<CHashLabels*> m_vpHashLabels;
};

template <typename Dtype>
class HashDataLayer : public Layer<Dtype>, public InternalThread {
 public:
  explicit HashDataLayer(const LayerParameter& param);
  virtual ~HashDataLayer();
  virtual void LayerSetUp(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top);
  // Data layers should be shared by multiple solvers in parallel
  virtual inline bool ShareInParallel() const { return true; }
  // Data layers have no bottoms, so reshaping is trivial.
  virtual void Reshape(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) {}

  virtual inline const char* type() const { return "HashData"; }
  virtual inline int ExactNumBottomBlobs() const { return 0; }
  virtual inline int MinTopBlobs() const { return 1; }
protected:	//for hash data
	void BatchHierHashes_2_blobs(const std::vector<CHierarchyHash *> &batch_hashes, const std::vector<Blob<Dtype>*>& top_blobs,
		const std::vector<std::vector<Blob<int>*>> &structure_blobs);
	void init_global_structure_blobs(std::vector<std::vector<Blob<int>*>> &structure_blobs);
public:
	//for debug
	//void save_blobs_to_hashFiles(const std::vector<Blob<Dtype>*>& top, const char *main_body);
 protected:
  void Next();
  bool Skip();

  virtual void Forward_cpu(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top);
  virtual void Forward_gpu(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top);
  virtual void Backward_cpu(const vector<Blob<Dtype>*>& top,
      const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) {}
  virtual void Backward_gpu(const vector<Blob<Dtype>*>& top,
      const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) {}
protected:
  void LoadHashLabelFileData(const char* filename);
  bool Load_labels(Blob<Dtype> &label_blob, FILE *fp);


  std::vector<std::string> hash_filenames_;
  unsigned int num_files_;
  unsigned int current_file_;
  unsigned int current_row_;
  Blob<Dtype>  label_blob_;
  std::vector<unsigned int> data_permutation_;
  std::vector<unsigned int> file_permutation_;
  uint64_t offset_;
  //hash
public:
	void destroyHierHashes();
	void destroyBatch();
	int loadHierHashes(std::vector<CHierarchyHash *> &vpHierHashes, FILE *fp);
  std::vector<CHierarchyHash *> m_vpHierHashes;
  //record current batch
  std::vector<CHierarchyHash *> m_hierHash_batch;

  //for multithread acceleration
public:
	virtual void InternalThreadEntry();
	void fetch_batch(HierHash_Label_HashLabel_Batch<Dtype>* batch);
	void PreLoadHashLabelFileData(std::vector<CHierarchyHash*> &vpHierHashes, Blob<Dtype> &label_blob, const char* filename);
	bool PreLoadHashDataHashLabels(std::vector<CHierarchyHash*> &vpHierHashes, std::vector<CHashLabels*> &vpHashLabels, const char *filename);
	void update_from_preload(HierHash_Label_HashLabel_Batch<Dtype>* prefetch);
public:
	vector<shared_ptr<HierHash_Label_HashLabel_Batch<Dtype> > > prefetch_;
	BlockingQueue<HierHash_Label_HashLabel_Batch<Dtype>*> prefetch_free_;
	BlockingQueue<HierHash_Label_HashLabel_Batch<Dtype>*> prefetch_full_;
	HierHash_Label_HashLabel_Batch<Dtype>* prefetch_current_;
public:	//for segmentation
	void destroyHashLabels();
	void destroyBatchHashLabels();
	bool loadHashDataHashLabels(const char *filename);
	bool load_hashLabels(std::vector<CHashLabels*> &vpHashLabels, FILE *fp);
	void BatchHashLabels_2_blobs(const std::vector<CHashLabels *> &vpHashLabels, Blob<Dtype>* top_label);
	void Next_from_prefetch();
	bool  segment_;	//if segment, the labels are packed together as hash data
	std::vector<CHashLabels*> m_vpHashLabels;
	std::vector<CHashLabels*> m_curBatchHashLabels;
};

}  // namespace caffe

#endif  // CAFFE_HDF5_DATA_LAYER_HPP_
