#ifndef CAFFE_UNPOOL_HASH_LAYER_HPP_
#define CAFFE_UNPOOL_HASH_LAYER_HPP_

#include <vector>

#include "caffe/blob.hpp"
#include "caffe/layer.hpp"
#include "caffe/proto/caffe.pb.h"
#include "caffe/util/HashData.h"

namespace caffe {

  /**
  * @brief Pools the input image by taking the max, average, etc. within regions.
  *
  * TODO(dox): thorough documentation for Forward, Backward, and proto params.
  */

  template <typename Dtype>
  class UnPoolHashLayer : public Layer<Dtype> {
  public:
    explicit UnPoolHashLayer(const LayerParameter& param)
      : Layer<Dtype>(param) {}
    virtual void LayerSetUp(const vector<Blob<Dtype>*>& bottom,
                            const vector<Blob<Dtype>*>& top);
    virtual void Reshape(const vector<Blob<Dtype>*>& bottom,
                         const vector<Blob<Dtype>*>& top);

    virtual inline const char* type() const { return "UnPoolHash"; }
    virtual inline int MinTopBlobs() const { return 1; }
    //// MAX POOL layers can output an extra top blob for the mask;
    //// others can only output the pooled inputs.
    //virtual inline int MaxTopBlobs() const {
    //  return (this->layer_param_.pooling_param().pool() ==
    //          PoolingParameter_PoolMethod_MAX) ? 2 : 1;
    //}
  protected:
	  void Forward_cpu_max(const vector<Blob<Dtype>*>& bottom,
		  const vector<Blob<Dtype>*>& top);
	  void forward_cpu_max(const Dtype *bottom_valid_data,
		  const CBatchStructure *bottom_struct,
		  Dtype *top_valid_data, const CBatchStructure *top_struct,
		  const Dtype *mask);
	  void Forward_gpu_max(const vector<Blob<Dtype>*>& bottom,
		  const vector<Blob<Dtype>*>& top);
	  void Backward_cpu_max(const vector<Blob<Dtype>*>& top,
		  const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom);
	  void backward_cpu_max(Dtype *bottom_valid_diff,
		  const CBatchStructure *bottom_struct,
		  const Dtype *top_valid_diff, const CBatchStructure *top_struct,
		  const Dtype *mask);
	  void Backward_gpu_max(const vector<Blob<Dtype>*>& top,
		  const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom);
  protected:
    virtual void Forward_cpu(const vector<Blob<Dtype>*>& bottom,
                             const vector<Blob<Dtype>*>& top);
    virtual void Forward_gpu(const vector<Blob<Dtype>*>& bottom,
                             const vector<Blob<Dtype>*>& top);
    virtual void Backward_cpu(const vector<Blob<Dtype>*>& top,
                              const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom);
    virtual void Backward_gpu(const vector<Blob<Dtype>*>& top,
                              const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom);

	Blob<int> kernel_shape_;
	Blob<int> stride_shape_;
	Blob<int> pad_shape_;
	
	int num_spatial_axes_;	//3 in 3D case
	int channels_;

	int m_bottom_struct_idx;
	int m_top_struct_idx;
  };

}  // namespace caffe

#endif  // CAFFE_POOLING_LAYER_HPP_
