#ifndef _HASH_DATA_H
#define _HASH_DATA_H

#include <vector>
#include "caffe/blob.hpp"
#include "MyMacro.h"
//used to pack xyz position 
typedef struct tag_PACKED_POSITION
{
	unsigned char _val[6];	//10 for X, 32 for Y, 54 for Z
}PACKED_POSITION;
typedef unsigned short VolumeIndexType;

inline void pack_xyz(PACKED_POSITION &packed_val, int x, int y, int z)
{
	//int &t0 = *(int*)(&packed_val._val[0]);
	//int &t1 = *(int*)(&packed_val._val[1]);
	//int &t2 = *(int*)(&packed_val._val[2]);
	//int &t3 = *(int*)(&packed_val._val[3]);
	//int &t4 = *(int*)(&packed_val._val[4]);
	//int &t5 = *(int*)(&packed_val._val[5]);
	//pack x y

	//NOTE!!!! (int*)(&val[0]] will occupy [0,1,2,3]. !!!!BUT!!!! 0 is the lowest position in the int!!!!
	//!!!THAT is, the int is ACTUALLY organized as 3210!!!!
	int &packed_xy = *(int*)(&packed_val._val[0]);
	packed_xy = ((packed_xy & 0xFFFF0000) | (x & 0x0000FFFF));
	packed_xy = ((packed_xy & 0x0000FFFF) | ((y & 0x0000FFFF) << 16));

	//pack z
	int &packed_z = *(int*)(&packed_val._val[2]);
	packed_z = (packed_z & 0x0000FFFF) | ((z & 0x0000FFFF) << 16);

}

inline void xyz_from_pack(const PACKED_POSITION &packed_val, int &x, int &y, int &z)
{
	const int &packed_xy = *(const int*)(&packed_val._val[0]);
	x = (packed_xy & 0x0000FFFF);
	y = (packed_xy & 0xFFFF0000) >> 16;

	const int &packed_z = *(const int*)(&packed_val._val[2]);
	z = (packed_z & 0xFFFF0000) >> 16;
}


//int getDefinedVoxelNum(const float *hash_data, int m_bar, int channels);
int getDefinedVoxelNum(const PACKED_POSITION *pos_tags, int m);
void getValidPoses(const PACKED_POSITION *pos_tags, int* valid_poses, int m);
//inline bool ishashVoxelDefined(const float *hash_voxel_ptr, int channels, int m)
//{
//	for (int c = 0; c < channels; c++)
//	{
//		if (*hash_voxel_ptr)
//		{
//			return true;
//			break;
//		}
//		hash_voxel_ptr+=m;
//	}
//	return false;
//}
inline bool ishashVoxelDefined(const PACKED_POSITION *pos_tag_ptr)
{
	int x, y, z;
	xyz_from_pack(*pos_tag_ptr, x, y, z);
	return !(x== INVALID_POSTAG || y== INVALID_POSTAG || z== INVALID_POSTAG);
}

void topMask_2_dense(const int *top_mask, const PACKED_POSITION *top_posTags, const unsigned char *top_offset,
	int top_m_bar, int top_r_bar, int channels, int top_res,
	const PACKED_POSITION *bottom_posTags, int bottom_res,
	int *dense_idx);

_inline int NXYZ2I(int nx, int ny, int nz, int n, int n2)
{
	return nz*n2 + ny*n + nx;
};

_inline void Hash(int nx, int ny, int nz, int& mx, int& my, int& mz,
	const unsigned char *offset_data, int m_bar, int r_bar, int r2)
{
	int rx = nx%r_bar;
	int ry = ny%r_bar;
	int rz = nz%r_bar;

	const unsigned char *offset = &offset_data[NXYZ2I(rx, ry, rz, r_bar, r2) * 3];

	mx = (nx + offset[0]) % m_bar;
	my = (ny + offset[1]) % m_bar;
	mz = (nz + offset[2]) % m_bar;
}



// when paralleling voxels * channels, we should banlance between hash computation with parallization
// so we should not parallel for all channels, we need to compute a group of channels instead.
#define CHANNEL_GROUP_NUM 32
#ifdef __CUDACC__
__device__ __host__ inline void xyz_from_pack_g(PACKED_POSITION packed_val, int &x, int &y, int &z)
{
	const int packed_xy = *(const int*)(packed_val._val);
	x = (packed_xy & 0x0000FFFF);
	y = (packed_xy & 0xFFFF0000) >> 16;

	const int packed_z = *(const int*)(packed_val._val + 2);
	z = (packed_z & 0xFFFF0000) >> 16;
}

__device__ __host__ inline int3 xyz_from_pack_g(PACKED_POSITION packed_val)
{
	int3 r;
	xyz_from_pack_g(packed_val, r.x, r.y, r.z);
	return r;
}

__device__ __host__ inline bool ishashVoxelDefined_g(int3 p)
{
	return !(p.x == INVALID_POSTAG || p.y == INVALID_POSTAG || p.z == INVALID_POSTAG);
}

__device__ __host__ inline bool ishashVoxelDefined_g(PACKED_POSITION pos_tag)
{
	return ishashVoxelDefined_g(xyz_from_pack_g(pos_tag));
}

__device__ __host__ inline int NXYZ2I_g(int nx, int ny, int nz, int n, int n2)
{
	return nz*n2 + ny*n + nx;
};

__device__ __host__ inline int NXYZ2I_g(int nx, int ny, int nz, int n)
{
	return (nz*n + ny)*n + nx;
};

__device__ __host__ inline int NXYZ2I_g(int3 p, int n)
{
	return NXYZ2I_g(p.x, p.y, p.z, n);
};

__device__ __host__ inline void I2NXYZ_g(int i, int& nx, int& ny, int& nz, int n, int n2)
{
	nz = i / n2; ny = (i - nz*n2) / n; nx = i - nz*n2 - ny*n;
};

__device__ __host__ inline void Hash_g(int nx, int ny, int nz, int& mx, int& my, int& mz,
	const unsigned char *offset_data, int m_bar, int r_bar, int r2)
{
	int rx = nx%r_bar;
	int ry = ny%r_bar;
	int rz = nz%r_bar;

	const unsigned char *offset = offset_data + NXYZ2I_g(rx, ry, rz, r_bar, r2) * 3;

	mx = (nx + offset[0]) % m_bar;
	my = (ny + offset[1]) % m_bar;
	mz = (nz + offset[2]) % m_bar;
}

__device__ __host__ inline int3 Hash_g(int nx, int ny, int nz,
	const unsigned char *offset_data, int m_bar, int r_bar)
{
	int3 r;
	Hash_g(nx, ny, nz, r.x, r.y, r.z, offset_data, m_bar, r_bar, r_bar*r_bar);
	return r;
}
#endif

/************************************************************************************************/
/*******************************For hierachy hashes**********************************/
/*****************************Different layers will have different hash structures********************/
/*****************************Different layers may share the same hash structures**************/

//hash struct info, may be shared by different layers
class CHashStructInfo	
{
public:
	CHashStructInfo();
	CHashStructInfo(const CHashStructInfo &rhs);
	~CHashStructInfo();
	CHashStructInfo &operator = (const CHashStructInfo &rhs);
	void destroy();
	int load(FILE *fp);
	int save(FILE *fp) const;
public:
	int *m_hash_index_table;
	PACKED_POSITION *m_position_tag;
	unsigned char *m_offset_data;
	int m_mBar;
	int m_rBar;
	int m_validNum;
	int m_res;
};

//preloaded hierhash
class CHierarchyHash
{
public:
	CHierarchyHash();
	CHierarchyHash(const CHierarchyHash &rhs);
	CHierarchyHash &operator = (const CHierarchyHash &rhs);
	~CHierarchyHash();
	void destroy();
	void destroyStructs();
	void initStructs(int n);
	int load(FILE *fp);
	int save(FILE *fp) const;
	int save(const char* filename) const;
public:
	float *m_valid_data;
	int m_channels;		//bottom channel
	std::vector<CHashStructInfo*> m_vpStructs;
};


int writeDense_2_TXT(const float *dense_data, int n, const char *filename);
int writeDense_2_TXT(const int *dense_data, int n, const char *filename);
int writeDense_2_TXT(const VolumeIndexType *dense_data, int n, const char *filename);
int writeDense_2_HF5(const float *dense_data, int n, int res, int channels, const char *filename);
int writeDense_2_BIN(const float *dense_data, int data_size, const char *filename);

int writeDense_2_Grid(const float *dense_data, int res, int channels, const char *filename);
int writeDense_2_Grid(const int *dense_data, int res, int channels, const char *filename);
/************************************UTILS********************************/
void calc_sum(const float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, float weight, float *out_weighted_sum);
void hash_add_scalar(float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, const float *to_adds);	
void hash_subtract_scalar(float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, const float *to_substracts);
void hash_mult_scalar(float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, const float *to_mults);
//used for calc variance
void calc_square_sum(const float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, float weight, float *out_weighted_sum);

/************************************************************************************/

/***********Labels in hash table format, for segmentation***********/
//hash labels for segmentation
class CHashLabels
{
public:
	CHashLabels();
	CHashLabels(const CHashLabels &rhs);
	~CHashLabels();
	CHashLabels &operator = (const CHashLabels &rhs);
	void destroy();
	int load(FILE *fp);
	int load(const char *filename);
	int save(FILE *fp) const;
public:
	float *m_pLabels;
	int m_m;
};


//for tranferring the parameters to functions
class CBatchStructure
{
public:
	CBatchStructure()
	{
		m_hashIdxTable_ptr = NULL;
		m_offset_ptr = NULL;
		m_posTag_ptr = NULL;
		m_volIdx_ptr = NULL;
		m_mBar_ptr = NULL;
		m_rBar_ptr = NULL;
		m_defNum_ptr = NULL;
		m_defNumSum_ptr = NULL;
		m_mSum_ptr = NULL;
		m_rSum_ptr = NULL;
		m_res_ptr = NULL;
		m_batch_size = 0;
	}
public:
	const int *m_hashIdxTable_ptr;
	const unsigned char *m_offset_ptr;
	const PACKED_POSITION *m_posTag_ptr;
	const VolumeIndexType *m_volIdx_ptr;
	const int* m_mBar_ptr;
	const int* m_rBar_ptr;
	const int* m_defNum_ptr;
	const int* m_defNumSum_ptr;
	const int *m_mSum_ptr;
	const int *m_rSum_ptr;
	const int *m_res_ptr;
	int m_batch_size;
};

//global data holder, storing the hash structures for the layers of CNN
template <typename Dtype>
class CGlobalSys
{
	CGlobalSys();
	~CGlobalSys();
public:
	static CGlobalSys *createInstance();
	static void destroyInstance();
	static CGlobalSys *get() { return s_globalSys; }
	static int isCreated() { return (int)(s_globalSys != NULL); }
public:
	static std::vector<std::vector<caffe::Blob<int>*>> &get_struct_blobs() { return s_globalSys->m_vvpStructBlobs; }
	static const caffe::Blob<int>* get_defNumSum_blob(int struct_idx);
	static caffe::Blob<Dtype> &get_workspace() { return s_globalSys->m_workspace; }
	static caffe::Blob<Dtype> &get_multiplier() { return s_globalSys->m_multiplier; }
	static void setBatchStructure(CBatchStructure &batch_structure, int struc_idx);
	static void setBatchStructure_GPU(CBatchStructure &batch_structure, int struc_idx);
protected:
	void init();
	void destroy();
protected:
	std::vector<std::vector<caffe::Blob<int>*>> m_vvpStructBlobs;
	caffe::Blob<Dtype> m_workspace;	//used for the intermediate matrix when convolution
	caffe::Blob<Dtype> m_multiplier;	//used for intermediate matrix multiplification
protected:
	static CGlobalSys<Dtype> *s_globalSys;
};

template<typename Dtype>
void batch_hash_2_dense(const Dtype *data, const CBatchStructure *self_struct,
	int channels,
	Dtype *dense_data);
template<typename Dtype>
void batch_dense_2_hash(Dtype *data, const CBatchStructure *self_struct,
	int channels,
	const Dtype *dense_data);

#endif