#ifndef _CAFFE_UTIL_HASH2COL_HPP_
#define _CAFFE_UTIL_HASH2COL_HPP_

#include "caffe/util/HashData.h"


namespace caffe {

	//NOTE: all matrix is row major
	template<typename Dtype>
	int hash2col_self_cpu(const Dtype* valid_data, int channels, const CBatchStructure *self_struct,
		const int kernel_shape[3],	//D, H, W
		Dtype* col_buff);

	template<typename Dtype>
	int hash2col_self_gpu(const Dtype* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		Dtype* col_buff);
#if RECORD_NEIGHBOR
	template<typename Dtype>
	int hash2col_self_gpu(const Dtype* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		Dtype* col_buff, int *nei_indices);
#endif

	//conver the col to bottom dif
	template<typename Dtype>
	int col2hash_self_cpu(Dtype* valid_data, int channels, const CBatchStructure *self_struct,
		const int kernel_shape[3],	//D, H, W
		const Dtype* col_buff);

	template<typename Dtype>
	int col2hash_self_gpu(Dtype* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const Dtype* col_buff);
#if RECORD_NEIGHBOR
	template<typename Dtype>
	int col2hash_self_gpu(Dtype* valid_data, int channels,
		int total_m, int total_def_num,
		const CBatchStructure *self_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const Dtype* col_buff, const int *nei_indices);
#endif


	//for convolution with stride>1
	//NOTE: all matrix is row major
	template<typename Dtype>
	int hash2col_ds_cpu(const Dtype* bt_data, int bt_channels, const CBatchStructure *bt_struct,
		const CBatchStructure *tp_struct,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		Dtype* col_buff);

	template<typename Dtype>
	int hash2col_ds_gpu(const Dtype* bt_data, int bt_channels,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		const CBatchStructure *bt_struct_GPU,
		const CBatchStructure *tp_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		Dtype* col_buff);


	template<typename Dtype>
	int col2hash_ds_cpu(Dtype* bt_data, int bt_channels, const CBatchStructure *bt_struct,
		const CBatchStructure *tp_struct,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		const Dtype* col_buff);

	template<typename Dtype>
	int col2hash_ds_gpu(Dtype* bt_data, int bt_channels,
		int tp_total_m, int tp_total_def_num, int bt_total_def_num,
		const CBatchStructure *bt_struct_GPU,
		const CBatchStructure *tp_struct_GPU,
		const int kernel_shape[3],	//D, H, W
		const int strides[3], const int pads[3],
		const Dtype* col_buff);
}  // namespace caffe

#endif  // CAFFE_UTIL_IM2COL_HPP_
