#include <vector>

#include "caffe/layers/deconv_hash_layer.hpp"

namespace caffe {

	template <typename Dtype>
	void DeconvHashLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
		const vector<Blob<Dtype>*>& top) 
	{
		const Dtype* weight = this->blobs_[0]->gpu_data();
		this->backward_gpu_gemm(bottom[HASH_DATA_BLOB]->gpu_data(), weight,
			top[HASH_DATA_BLOB]->mutable_gpu_data());


		if (this->bias_term_)
		{
			const Dtype* bias = this->blobs_[1]->gpu_data();
			this->forward_gpu_bias(top[HASH_DATA_BLOB]->mutable_gpu_data(), bias);
		}
	}

	template <typename Dtype>
	void DeconvHashLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
		const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom)
	{
		const Dtype* weight = this->blobs_[0]->gpu_data();
		Dtype* weight_diff = this->blobs_[0]->mutable_gpu_diff();
		
		const Dtype* top_diff = top[HASH_DATA_BLOB]->gpu_diff();
		// Bias gradient, if necessary.
		if (this->bias_term_ && this->param_propagate_down_[1]) 
		{
			Dtype* bias_diff = this->blobs_[1]->mutable_gpu_diff();
			this->backward_gpu_bias(bias_diff, top_diff);
		}
		if (this->param_propagate_down_[0] || propagate_down[HASH_DATA_BLOB]) 
		{
			const Dtype* bottom_data = bottom[HASH_DATA_BLOB]->gpu_data();
			Dtype* bottom_diff = bottom[HASH_DATA_BLOB]->mutable_gpu_diff();
			// gradient w.r.t. weight. Note that we will accumulate diffs.
			if (this->param_propagate_down_[0]) 
			{
				this->weight_gpu_gemm(top_diff,
					bottom_data, weight_diff);
			}
			// gradient w.r.t. bottom data, if necessary.
			if (propagate_down[HASH_DATA_BLOB]) 
			{
				this->forward_gpu_gemm(top_diff, weight,
					bottom_diff);
			}
		}
	}

	INSTANTIATE_LAYER_GPU_FUNCS(DeconvHashLayer);
}  // namespace caffe
