#include "caffe/util/HashData.h"
#include <string>
#include <fstream>
#include <sstream>
#include <cstdio>
#include "caffe/util/MyMacro.h"
#include "caffe/util/hdf5.hpp"

//#pragma comment(lib, "hdf5_hl.lib")
//#pragma comment(lib, "hdf5.lib")

//NOTE: the data is organized as channel * depth * height * weight
//int getDefinedVoxelNum(const float *hash_data, int m_bar, int channels)
//{
//	const int m = m_bar * m_bar * m_bar;
//	int defined_num = 0;
//	const float *data_ptr = hash_data;
//	for (int v = 0; v < m; v++)
//	{
//		if (ishashVoxelDefined(data_ptr, channels, m))
//		{
//			defined_num++;
//		}
//		data_ptr++;
//	}
//	return defined_num;
//}
int getDefinedVoxelNum(const PACKED_POSITION *pos_tags, int m)
{
	int defined_num = 0;
	const PACKED_POSITION *pos_tag_ptr = pos_tags;
	for (int v = 0; v < m; v++)
	{
		if (ishashVoxelDefined(pos_tag_ptr))
		{
			defined_num++;
		}
		pos_tag_ptr++;
	}
	return defined_num;
}

void getValidPoses(const PACKED_POSITION *pos_tags, int* valid_poses, int m)
{
	const PACKED_POSITION *pos_tag_ptr = pos_tags;
	for (int v = 0; v < m; v++)
	{
		if (ishashVoxelDefined(pos_tag_ptr))
		{
			*valid_poses++ = v;
		}
		pos_tag_ptr++;
	}
}


template<typename Dtype>
void batch_hash_2_dense(const Dtype *data, const CBatchStructure *self_struct,
	int channels,
	Dtype *dense_data)
{
	const int batch_size = self_struct->m_batch_size;

	const Dtype *cur_valid_data = data;
	Dtype *cur_dense_data = dense_data;

	const PACKED_POSITION *position_tags = self_struct->m_posTag_ptr;
	const int *hash_index_table = self_struct->m_hashIdxTable_ptr;

	const int total_defNum = self_struct->m_defNumSum_ptr[batch_size];

	for (int i = 0; i < batch_size; ++i)
	{
		const int m_bar = self_struct->m_mBar_ptr[i];
		const int m = m_bar * m_bar * m_bar;
		const int defNum = self_struct->m_defNum_ptr[i];
		const int res = self_struct->m_res_ptr[i];
		const int res3 = res * res * res;
		for (int v = 0; v < m; v++)
		{
			if (!ishashVoxelDefined(&position_tags[v]))
			{
				continue;
			}
			const int valid_idx = hash_index_table[v];
			if (valid_idx < 0 || valid_idx >= defNum)
			{
				printf("Fatal error: valid idx out of range!\n");
				exit(0);
			}

			//get the real voxel position from the position tag
			int cx, cy, cz;
			xyz_from_pack(position_tags[v], cx, cy, cz);	
			int ni = NXYZ2I(cx, cy, cz, res, res*res);
			
			Dtype *cur_dense_ptr = cur_dense_data + ni;
			const Dtype *cur_valid_ptr = cur_valid_data + valid_idx;
			for (int c = 0; c < channels; c++)
			{
				*cur_dense_ptr = *cur_valid_ptr;
				cur_dense_ptr += res3;
				cur_valid_ptr += total_defNum;
			}
		}

		//to next hash
		position_tags += m;
		hash_index_table += m;
		cur_valid_data += defNum;
		//next dense
		cur_dense_data += channels * res3;
	}
}


template<typename Dtype>
void batch_dense_2_hash(Dtype *data, const CBatchStructure *self_struct,
	int channels,
	const Dtype *dense_data)
{
	const int batch_size = self_struct->m_batch_size;

	Dtype *cur_valid_data = data;
	const Dtype *cur_dense_data = dense_data;

	const PACKED_POSITION *position_tags = self_struct->m_posTag_ptr;
	const int *hash_index_table = self_struct->m_hashIdxTable_ptr;

	const int total_defNum = self_struct->m_defNumSum_ptr[batch_size];

	for (int i = 0; i < batch_size; ++i)
	{
		const int m_bar = self_struct->m_mBar_ptr[i];
		const int m = m_bar * m_bar * m_bar;
		const int defNum = self_struct->m_defNum_ptr[i];
		const int res = self_struct->m_res_ptr[i];
		const int res3 = res * res * res;
		for (int v = 0; v < m; v++)
		{
			if (!ishashVoxelDefined(&position_tags[v]))
			{
				continue;
			}
			const int valid_idx = hash_index_table[v];
			if (valid_idx < 0 || valid_idx >= defNum)
			{
				printf("Fatal error: valid idx out of range!\n");
				exit(0);
			}

			//get the real voxel position from the position tag
			int cx, cy, cz;
			xyz_from_pack(position_tags[v], cx, cy, cz);
			int ni = NXYZ2I(cx, cy, cz, res, res*res);

			const Dtype *cur_dense_ptr = cur_dense_data + ni;
			Dtype *cur_valid_ptr = cur_valid_data + valid_idx;
			for (int c = 0; c < channels; c++)
			{
				*cur_valid_ptr = *cur_dense_ptr;
				cur_dense_ptr += res3;
				cur_valid_ptr += total_defNum;
			}
		}

		//to next hash
		position_tags += m;
		hash_index_table += m;
		cur_valid_data += defNum;
		//next dense
		cur_dense_data += channels * res3;
	}
}



void topMask_2_dense(const int *top_mask, const PACKED_POSITION *top_posTags, const unsigned char *top_offset,
	int top_m_bar, int top_r_bar, int channels, int top_res,
	const PACKED_POSITION *bottom_posTags, int bottom_res,
	int *dense_idx)
{
	int top_res3 = top_res*top_res*top_res;
	memset(dense_idx, -1, sizeof(int)*top_res3 * channels);

	int top_m = top_m_bar * top_m_bar * top_m_bar;
	const int *mask_ptr = top_mask;
	for (int i = 0; i < top_m; i++)
	{
		if (!ishashVoxelDefined(&top_posTags[i]))
		{
			mask_ptr++;
			continue;
		}
		int x, y, z;
		xyz_from_pack(top_posTags[i], x, y, z);
		int ni = NXYZ2I(x, y, z, top_res, top_res*top_res);
		int *cur_dense_ptr = dense_idx + ni;
		const int *cur_mask_ptr = mask_ptr;
		for (int c = 0; c < channels; c++)
		{
			int bt_hash_idx = *cur_mask_ptr;
			if (bt_hash_idx < 0)
			{
				printf("Fatal error: impossible!\n");
				exit(0);
			}
			int bx, by, bz;
			xyz_from_pack(bottom_posTags[bt_hash_idx], bx, by, bz);
			*cur_dense_ptr = NXYZ2I(bx, by, bz, bottom_res, bottom_res*bottom_res);
			
			cur_dense_ptr += top_res3;
			cur_mask_ptr += top_m;
		}
		mask_ptr++;
	}
}


/*********************************************************************************************/
/**********************************Hierarchy hashes*******************************************/
CHashStructInfo::CHashStructInfo()
{
	m_hash_index_table = NULL;
	m_position_tag = NULL;
	m_offset_data = NULL;
	m_mBar = 0;
	m_rBar = 0;
	m_validNum = 0;
	m_res = 0;
}


CHashStructInfo::CHashStructInfo(const CHashStructInfo &rhs)
{
	m_mBar = rhs.m_mBar;
	m_rBar = rhs.m_rBar;
	m_validNum = rhs.m_validNum;
	m_res = rhs.m_res;

	const int m = m_mBar*m_mBar*m_mBar;
	const int r = m_rBar * m_rBar * m_rBar;

	m_offset_data = new unsigned char[r * 3];
	m_position_tag = new PACKED_POSITION[m];
	m_hash_index_table = new int[m];
	
	memcpy(m_offset_data, rhs.m_offset_data, sizeof(unsigned char) * r * 3);
	memcpy(m_position_tag, rhs.m_position_tag, sizeof(PACKED_POSITION) * m);
	memcpy(m_hash_index_table, rhs.m_hash_index_table, sizeof(int) * m);
}


CHashStructInfo &CHashStructInfo::operator=(const CHashStructInfo &rhs)
{
	const int ori_mBar = m_mBar;
	const int ori_rBar = m_rBar;
	const int ori_m = ori_mBar * ori_mBar * ori_mBar;
	const int ori_r = ori_rBar * ori_rBar * ori_rBar;
	
	m_mBar = rhs.m_mBar;
	m_rBar = rhs.m_rBar;
	m_validNum = rhs.m_validNum;
	m_res = rhs.m_res;

	const int m = m_mBar*m_mBar*m_mBar;
	const int r = m_rBar * m_rBar * m_rBar;

	if (r > ori_r)
	{
		SAFE_VDELETE(m_offset_data);
		m_offset_data = new unsigned char[r * 3];
	}
	if (m > ori_m)
	{
		SAFE_VDELETE(m_position_tag);
		m_position_tag = new PACKED_POSITION[m];
		SAFE_VDELETE(m_hash_index_table);
		m_hash_index_table = new int[m];
	}

	memcpy(m_offset_data, rhs.m_offset_data, sizeof(unsigned char) * r * 3);
	memcpy(m_position_tag, rhs.m_position_tag, sizeof(PACKED_POSITION) * m);
	memcpy(m_hash_index_table, rhs.m_hash_index_table, sizeof(int) * m);
	return *this;
}

CHashStructInfo::~CHashStructInfo()
{
	destroy();
}

void CHashStructInfo::destroy()
{
	SAFE_VDELETE(m_position_tag);	
	SAFE_VDELETE(m_offset_data);
	SAFE_VDELETE(m_hash_index_table);
	m_mBar = 0;
	m_rBar = 0;
	m_validNum = 0;
	m_res = 0;
}

int CHashStructInfo::save(FILE *fp) const
{
	fwrite(&m_res, sizeof(int), 1, fp);
	fwrite(&m_validNum, sizeof(int), 1, fp);
	fwrite(&m_mBar, sizeof(int), 1, fp);	//m, for hash table and position hash table
	fwrite(&m_rBar, sizeof(int), 1, fp);	//r, for offset table
											//hash index table
	const int m = m_mBar * m_mBar * m_mBar;
	const int r = m_rBar * m_rBar * m_rBar;
	fwrite(m_hash_index_table, sizeof(int), m, fp);
	fwrite(m_offset_data, sizeof(unsigned char), r * 3, fp);
	fwrite(m_position_tag, sizeof(PACKED_POSITION), m, fp);
	return true;
}

int CHashStructInfo::load(FILE *fp)
{
	const int ori_mBar = m_mBar;
	const int ori_rBar = m_rBar;
	const int ori_m = ori_mBar * ori_mBar * ori_mBar;
	const int ori_r = ori_rBar * ori_rBar * ori_rBar;

	fread(&m_res, sizeof(int), 1, fp);
	fread(&m_validNum, sizeof(int), 1, fp);
	fread(&m_mBar, sizeof(int), 1, fp);	//m, for hash table and position hash table
	fread(&m_rBar, sizeof(int), 1, fp);	//r, for offset table

	const int m = m_mBar*m_mBar*m_mBar;
	const int r = m_rBar * m_rBar * m_rBar;

	if (r <= 0 || m <= 0)
	{
		printf("Error: failed to load hash struct info! invalid r_bar m_bar!\n");
		return 0;
	}

	if (r > ori_r)
	{
		SAFE_VDELETE(m_offset_data);
		m_offset_data = new unsigned char[r * 3];
	}

	if (m > ori_m)
	{
		SAFE_VDELETE(m_position_tag);
		m_position_tag = new PACKED_POSITION[m];
		SAFE_VDELETE(m_hash_index_table);
		m_hash_index_table = new int[m];
	}
	fread(m_hash_index_table, sizeof(int), m, fp);
	fread(m_offset_data, sizeof(unsigned char), r * 3, fp);
	fread(m_position_tag, sizeof(PACKED_POSITION), m, fp);
	return 1;
}

///////////////////////////////////////////////////////////////////
CHierarchyHash::CHierarchyHash()
{
	m_valid_data = NULL;
	m_channels = 0;
}

CHierarchyHash::CHierarchyHash(const CHierarchyHash &rhs)
{
	m_channels = rhs.m_channels;
	m_vpStructs.resize(rhs.m_vpStructs.size());
	for (int i=0;i<(int)rhs.m_vpStructs.size();i++)
	{
		m_vpStructs[i] = new CHashStructInfo(*rhs.m_vpStructs[i]);
	}

	if (m_vpStructs.size())
	{
		const int bottom_valid_num = m_vpStructs[0]->m_validNum;
		m_valid_data = new float[bottom_valid_num*m_channels];
		memcpy(m_valid_data, rhs.m_valid_data, sizeof(float)*bottom_valid_num*m_channels);
	}
	else
	{
		m_valid_data = NULL;
	}

}

CHierarchyHash::~CHierarchyHash()
{
	destroy();
}

CHierarchyHash &CHierarchyHash::operator = (const CHierarchyHash &rhs)
{
	//record ori
	int ori_channels = m_channels;
	int ori_validNum = 0;
	if (m_vpStructs.size())
	{
		ori_validNum = m_vpStructs[0]->m_validNum;
	}

	//new val
	m_channels = rhs.m_channels;

	if (!rhs.m_vpStructs.size())
	{
		printf("Fatal error when CHierarchyHash::operator=(); no structure info!\n");
		exit(0);
	}
	if (m_vpStructs.size() && m_vpStructs.size() != rhs.m_vpStructs.size())
	{
		printf("FATAL error: structure number changed!! UNEXPECTED!!!\n");
		exit(0);
	}

	if (!m_vpStructs.size())
	{
		initStructs((int)rhs.m_vpStructs.size());
	}
	//copy structs
	for (int i = 0; i<(int)rhs.m_vpStructs.size(); i++)
	{
		*m_vpStructs[i] = *rhs.m_vpStructs[i];
	}

	//copy data
	const int new_validNum = m_vpStructs[0]->m_validNum;
	if (new_validNum * m_channels > ori_validNum * ori_channels)	//need larger memory
	{
		SAFE_VDELETE(m_valid_data);
		m_valid_data = new float[new_validNum*m_channels];
	}
	memcpy(m_valid_data, rhs.m_valid_data, sizeof(float)*new_validNum*m_channels);
	return *this;
}

void CHierarchyHash::destroy()
{
	SAFE_VDELETE(m_valid_data);
	m_channels = 0;
	destroyStructs();
}

void CHierarchyHash::destroyStructs()
{
	for (int i = 0; i < (int)m_vpStructs.size(); i++)
	{
		delete m_vpStructs[i];
	}
	m_vpStructs.resize(0);
}

void CHierarchyHash::initStructs(int n)
{
	destroyStructs();
	m_vpStructs.resize(n);
	for (int i = 0; i < n; i++)
	{
		m_vpStructs[i] = new CHashStructInfo();
	}
}

int CHierarchyHash::load(FILE *fp)
{
	int ori_channels = m_channels;
	int ori_valid_num = 0;
	if (m_vpStructs.size())
	{
		ori_valid_num = m_vpStructs[0]->m_validNum;
	}

	//first load hash structures
	int structure_num;
	fread(&structure_num, sizeof(int), 1, fp);

	if (structure_num < 1)
	{
		printf("Fatal error when CHierarchyHash::load; no structure info!\n");
		exit(0);
	}

	if (m_vpStructs.size() && m_vpStructs.size() != structure_num)
	{
		printf("FATAL error: structure number changed!! UNEXPECTED!!!\n");
		exit(0);
	}

	if (!m_vpStructs.size())
	{
		initStructs(structure_num);
	}
	for (int i = 0; i<structure_num; i++)
	{
		m_vpStructs[i]->load(fp);
	}

	//read bottom hash data
	const int new_valid_num = m_vpStructs[0]->m_validNum;

	fread(&m_channels, sizeof(int), 1, fp);
	if (new_valid_num * m_channels > ori_valid_num * ori_channels)	//need larger memory
	{
		SAFE_VDELETE(m_valid_data);
		m_valid_data = new float[new_valid_num*m_channels];
	}
	fread(m_valid_data, sizeof(float), new_valid_num*m_channels, fp);

	return 1;
}

int CHierarchyHash::save(FILE *fp) const 
{
	int structure_num = (int)m_vpStructs.size();
	fwrite(&structure_num, sizeof(int), 1, fp);

	if (structure_num < 1)
	{
		printf("Fatal error when CHierarchyHash::save; no structure info!\n");
		exit(0);
	}

	for (int i = 0; i < structure_num; i++)
	{
		m_vpStructs[i]->save(fp);
	}

	const int valid_num = m_vpStructs[0]->m_validNum;
	fwrite(&m_channels, sizeof(int), 1, fp);
	fwrite(m_valid_data, sizeof(float), valid_num*m_channels, fp);

	return 1;
}

int CHierarchyHash::save(const char* filename) const
{
	FILE *fp = fopen(filename, "wb");
	if (!fp)
	{
		printf("Error: failed to save hier hash to %s\n", filename);
		return false;
	}
	int flag = save(fp);
	fclose(fp);
	return flag;
}

/*****************************************************/

int writeDense_2_HF5(const float *dense_data, int n, int res, int channels, const char *filename)
{
	hid_t file_id = H5Fcreate(filename, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
	if (filename < 0)
	{
		printf("Error: write dense 2 HF5 file %s failed!\n", filename);
		return 0;
	}

	herr_t err = 0;
	hsize_t dims[5] = { (hsize_t)n, (hsize_t)channels, (hsize_t)res, (hsize_t)res, (hsize_t)res };

	err = H5LTmake_dataset_float(file_id, "volume", 5, dims, dense_data);

	err = H5Fclose(file_id);

	return 1;
}

int writeDense_2_TXT(const float *dense_data, int n, const char *filename)
{
	std::ofstream ofs(filename);
	for (int i = 0; i < n; i++)
	{
		ofs << dense_data[i] << std::endl;
	}
	ofs.close();
	return 1;
}

int writeDense_2_TXT(const int *dense_data, int n, const char *filename)
{
	std::ofstream ofs(filename);
	for (int i = 0; i < n; i++)
	{
		ofs << dense_data[i] << std::endl;
	}
	ofs.close();
	return 1;
}

int writeDense_2_TXT(const VolumeIndexType *dense_data, int n, const char *filename)
{
	std::ofstream ofs(filename);
	for (int i = 0; i < n; i++)
	{
		ofs << dense_data[i] << std::endl;
	}
	ofs.close();
	return 1;
}

int writeDense_2_BIN(const float *dense_data, int data_size, const char *filename)
{
	FILE *fp = fopen(filename, "wb");
	if (!fp)
	{
		printf("Error: failed to write dense to %s\n",filename);
		return 0;
	}
	
	fwrite(dense_data,sizeof(float),data_size,fp);

	fclose(fp);
	return 1;
}



int writeDense_2_Grid(const float *dense_data, int res, int channels, const char *filename)
{
	FILE *fp = fopen(filename, "wb");
	if (!fp)
	{
		printf("Error: failed to load dense from %s\n", filename);
		return 0;
	}
	fwrite(&channels, sizeof(int), 1, fp);
	fwrite(&res, sizeof(int), 1, fp);
	fwrite(&res, sizeof(int), 1, fp);
	fwrite(&res, sizeof(int), 1, fp);
	
	//fwrite(dense_data, sizeof(float), channels*res*res*res, fp);
	//NOTE: the grid memory order is D H W C
	int tn = res*res*res;
	const float *data_ptr = dense_data;
	for (int i=0;i<tn;i++)
	{
		const float *cur_ptr = data_ptr;
		for (int c=0;c<channels;c++)
		{
			fwrite(cur_ptr,sizeof(float),1,fp);
			cur_ptr += tn;
		}

		data_ptr++;
	}
	
	fclose(fp);
	return 1;
}



int writeDense_2_Grid(const int *dense_data, int res, int channels, const char *filename)
{
	FILE *fp = fopen(filename, "wb");
	if (!fp)
	{
		printf("Error: failed to load dense from %s\n", filename);
		return 0;
	}
	fwrite(&channels, sizeof(int), 1, fp);
	fwrite(&res, sizeof(int), 1, fp);
	fwrite(&res, sizeof(int), 1, fp);
	fwrite(&res, sizeof(int), 1, fp);

	//fwrite(dense_data, sizeof(float), channels*res*res*res, fp);
	//NOTE: the grid memory order is D H W C
	int tn = res*res*res;
	const int *data_ptr = dense_data;
	for (int i = 0; i < tn; i++)
	{
		const int *cur_ptr = data_ptr;
		for (int c = 0; c < channels; c++)
		{
			fwrite(cur_ptr, sizeof(float), 1, fp);
			cur_ptr += tn;
		}

		data_ptr++;
	}

	fclose(fp);
	return 1;
}



/***********************UTILS*************************/
void calc_sum(const float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, float weight, float *out_weighted_sum)
{
	const int m = m_bar * m_bar * m_bar;
	
	memset(out_weighted_sum, 0, sizeof(float)*channels);

	for (int v = 0; v < m; v++)
	{
		//if the hash voxel is undefined, skip
		if (!ishashVoxelDefined(&posTag[v]))
		{
			continue;
		}
		///////////////////////////////////////////

		const float *hash_ptr = &hash[v];
		for (int c = 0; c < channels; c++)
		{
			out_weighted_sum[c] += (*hash_ptr) * weight;
			hash_ptr += m;
		}

	}
}


void hash_add_scalar(float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, const float *to_adds)
{
	const int m = m_bar * m_bar * m_bar;
	
	for (int v = 0; v < m; v++)
	{
		//if the hash voxel is undefined, skip
		if (!ishashVoxelDefined(&posTag[v]))
		{
			continue;
		}
		///////////////////////////////////////////

		float *hash_ptr = &hash[v];
		for (int c = 0; c < channels; c++)
		{
			*hash_ptr += to_adds[c];
			hash_ptr += m;
		}
	}
}


void hash_subtract_scalar(float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, const float *to_subtracts)
{
	const int m = m_bar * m_bar * m_bar;

	for (int v = 0; v < m; v++)
	{
		//if the hash voxel is undefined, skip
		if (!ishashVoxelDefined(&posTag[v]))
		{
			continue;
		}
		///////////////////////////////////////////

		float *hash_ptr = &hash[v];
		for (int c = 0; c < channels; c++)
		{
			*hash_ptr -= to_subtracts[c];
			hash_ptr += m;
		}
	}
}

void hash_mult_scalar(float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, const float *to_mults)
{
	const int m = m_bar * m_bar * m_bar;

	for (int v = 0; v < m; v++)
	{
		//if the hash voxel is undefined, skip
		if (!ishashVoxelDefined(&posTag[v]))
		{
			continue;
		}
		///////////////////////////////////////////

		float *hash_ptr = &hash[v];
		for (int c = 0; c < channels; c++)
		{
			*hash_ptr *= to_mults[c];
			hash_ptr += m;
		}
	}
}


void calc_square_sum(const float *hash, const unsigned char *offset,
	const PACKED_POSITION *posTag, int m_bar, int r_bar,
	int channels, int def_num, float weight, float *out_weighted_sum)
{
	const int m = m_bar * m_bar * m_bar;

	memset(out_weighted_sum, 0, sizeof(float)*channels);

	for (int v = 0; v < m; v++)
	{
		//if the hash voxel is undefined, skip
		if (!ishashVoxelDefined(&posTag[v]))
		{
			continue;
		}
		///////////////////////////////////////////

		const float *hash_ptr = &hash[v];
		for (int c = 0; c < channels; c++)
		{
			out_weighted_sum[c] += (*hash_ptr) * (*hash_ptr) * weight;
			hash_ptr += m;
		}

	}
}

/**********************************Hash labels**************************************/
CHashLabels::CHashLabels()
{
	m_pLabels = NULL;
	m_m = 0;
}

CHashLabels::CHashLabels(const CHashLabels &rhs)
{
	m_m = rhs.m_m;
	m_pLabels = new float[m_m];
	memcpy(m_pLabels,rhs.m_pLabels,sizeof(float)*m_m);
}

CHashLabels::~CHashLabels()
{
	destroy();
}

CHashLabels &CHashLabels::operator = (const CHashLabels &rhs)
{
	int ori_m_m = m_m;

	m_m = rhs.m_m;

	if (m_m > ori_m_m)
	{
		SAFE_VDELETE(m_pLabels);
		m_pLabels = new float[m_m];
	}
	memcpy(m_pLabels, rhs.m_pLabels, sizeof(float)*m_m);

	return *this;
}

void CHashLabels::destroy()
{
	SAFE_VDELETE(m_pLabels);
	m_m = 0;
}

int CHashLabels::load(FILE *fp)
{
	int ori_m = m_m;
	fread(&m_m, sizeof(int), 1, fp);
	if (m_m > ori_m)
	{
		SAFE_VDELETE(m_pLabels);
		m_pLabels = new float[m_m];
	}
	
	fread(m_pLabels, sizeof(float), m_m, fp);

	return 1;
}

int CHashLabels::save(FILE *fp) const
{
	fwrite(&m_m, sizeof(int), 1, fp);
	fwrite(m_pLabels, sizeof(float), m_m, fp);
	return 1;
}

int CHashLabels::load(const char *filename)
{
	FILE *fp = fopen(filename, "rb");
	if (!fp)
	{
		printf("Error: failed to read PSH labels from %s\n", filename);
		return false;
	}
	int flag = load(fp);
	fclose(fp);
	return flag;
}


/*********************************Global data holder*************************************/
//global data holder, storing the hash structures for the layers of CNN
template<typename Dtype>
CGlobalSys<Dtype>::CGlobalSys()
{
	init();
}

template<typename Dtype>
CGlobalSys<Dtype>::~CGlobalSys()
{
	destroy();
}

template<typename Dtype>
CGlobalSys<Dtype> *CGlobalSys<Dtype>::createInstance()
{
	SAFE_DELETE(s_globalSys);
	s_globalSys = new CGlobalSys<Dtype>();
	return s_globalSys;
}

template<typename Dtype>
void CGlobalSys<Dtype>::destroyInstance()
{
	SAFE_DELETE(s_globalSys);
}

template<typename Dtype>
void CGlobalSys<Dtype>::init()
{

}

template<typename Dtype>
void CGlobalSys<Dtype>::destroy()
{
	for (int i = 0; i < (int)m_vvpStructBlobs.size(); i++)
	{
		for (int j = 0; j < (int)m_vvpStructBlobs[i].size(); j++)
		{
			delete m_vvpStructBlobs[i][j];
		}
	}
	m_vvpStructBlobs.resize(0);
}

template<typename Dtype>
const caffe::Blob<int>* CGlobalSys<Dtype>::get_defNumSum_blob(int struct_idx)
{
	if (struct_idx < 0 || struct_idx >= (int)s_globalSys->m_vvpStructBlobs.size())
	{
		printf("Fatal error: struct idx %d out of range [0, %d)\n",struct_idx, (int)s_globalSys->m_vvpStructBlobs.size());
		exit(0);
	}
	return s_globalSys->m_vvpStructBlobs[struct_idx][DEFNUM_SUM_BLOB];
}

template<typename Dtype>
void CGlobalSys<Dtype>::setBatchStructure(CBatchStructure &batch_structure, int struct_idx)
{
	if (struct_idx < 0 || struct_idx >= (int)s_globalSys->m_vvpStructBlobs.size())
	{
		printf("Fatal error: struct idx %d out of range [0, %d)\n", struct_idx, (int)s_globalSys->m_vvpStructBlobs.size());
		exit(0);
	}
	const std::vector<caffe::Blob<int>*> &oneStruct_blobs = s_globalSys->m_vvpStructBlobs[struct_idx];

	batch_structure.m_hashIdxTable_ptr = oneStruct_blobs[HASH_TABLE_BLOB]->cpu_data();
	batch_structure.m_offset_ptr = (const unsigned char*)oneStruct_blobs[OFFSET_BLOB]->cpu_data();
	batch_structure.m_posTag_ptr = (const PACKED_POSITION*)oneStruct_blobs[POSTAG_BLOB]->cpu_data();
	batch_structure.m_volIdx_ptr = (const VolumeIndexType*)oneStruct_blobs[VOLUME_IDX_BLOB]->cpu_data();
	batch_structure.m_mBar_ptr = oneStruct_blobs[M_BAR_BLOB]->cpu_data();
	batch_structure.m_rBar_ptr = oneStruct_blobs[R_BAR_BLOB]->cpu_data();
	batch_structure.m_defNum_ptr = oneStruct_blobs[DEFNUM_BLOB]->cpu_data();
	batch_structure.m_defNumSum_ptr = oneStruct_blobs[DEFNUM_SUM_BLOB]->cpu_data();
	batch_structure.m_mSum_ptr = oneStruct_blobs[M_SUM_BLOB]->cpu_data();
	batch_structure.m_rSum_ptr = oneStruct_blobs[R_SUM_BLOB]->cpu_data();
	batch_structure.m_res_ptr = oneStruct_blobs[RESOLUTION_BLOB]->cpu_data();
	batch_structure.m_batch_size = oneStruct_blobs[M_BAR_BLOB]->count();
}

template<typename Dtype>
void CGlobalSys<Dtype>::setBatchStructure_GPU(CBatchStructure &batch_structure, int struct_idx)
{
	if (struct_idx < 0 || struct_idx >= (int)s_globalSys->m_vvpStructBlobs.size())
	{
		printf("Fatal error: struct idx %d out of range [0, %d)\n", struct_idx, (int)s_globalSys->m_vvpStructBlobs.size());
		exit(0);
	}
	const std::vector<caffe::Blob<int>*> &oneStruct_blobs = s_globalSys->m_vvpStructBlobs[struct_idx];

	batch_structure.m_hashIdxTable_ptr = oneStruct_blobs[HASH_TABLE_BLOB]->gpu_data();
	batch_structure.m_offset_ptr = (const unsigned char*)oneStruct_blobs[OFFSET_BLOB]->gpu_data();
	batch_structure.m_posTag_ptr = (const PACKED_POSITION*)oneStruct_blobs[POSTAG_BLOB]->gpu_data();
	batch_structure.m_volIdx_ptr = (const VolumeIndexType*)oneStruct_blobs[VOLUME_IDX_BLOB]->gpu_data();
	batch_structure.m_mBar_ptr = oneStruct_blobs[M_BAR_BLOB]->gpu_data();
	batch_structure.m_rBar_ptr = oneStruct_blobs[R_BAR_BLOB]->gpu_data();
	batch_structure.m_defNum_ptr = oneStruct_blobs[DEFNUM_BLOB]->gpu_data();
	batch_structure.m_defNumSum_ptr = oneStruct_blobs[DEFNUM_SUM_BLOB]->gpu_data();
	batch_structure.m_mSum_ptr = oneStruct_blobs[M_SUM_BLOB]->gpu_data();
	batch_structure.m_rSum_ptr = oneStruct_blobs[R_SUM_BLOB]->gpu_data();
	batch_structure.m_res_ptr = oneStruct_blobs[RESOLUTION_BLOB]->gpu_data();
	batch_structure.m_batch_size = oneStruct_blobs[M_BAR_BLOB]->count();
}



template<> CGlobalSys<float>* CGlobalSys<float>::s_globalSys = NULL;    
template class CGlobalSys<float>;
template<> CGlobalSys<double>* CGlobalSys<double>::s_globalSys = NULL;
template class CGlobalSys<double>;


template void batch_hash_2_dense<float>(const float *data, const CBatchStructure *self_struct,
	int channels, float *dense_data);
template void batch_hash_2_dense<double>(const double *data, const CBatchStructure *self_struct,
	int channels, double *dense_data);

template void batch_dense_2_hash<float>(float *data, const CBatchStructure *self_struct,
	int channels, const float *dense_data);
template void batch_dense_2_hash<double>(double *data, const CBatchStructure *self_struct,
	int channels, const double *dense_data);